"""
Base storage interface.

All storage backends must implement this interface.
"""

from abc import ABC, abstractmethod
from typing import List, Optional


class BaseStore(ABC):

    @abstractmethod
    def put_note(self, note: dict) -> str:
        pass

    @abstractmethod
    def get_note(self, note_id: str) -> Optional[dict]:
        pass

    @abstractmethod
    def list_notes(self, limit: int = 100, offset: int = 0, agent_id: str = None) -> List[dict]:
        pass

    @abstractmethod
    def delete_note(self, note_id: str) -> bool:
        pass

    @abstractmethod
    def put_cell(self, cell: dict) -> str:
        pass

    @abstractmethod
    def get_cell(self, cell_id: str) -> Optional[dict]:
        pass

    @abstractmethod
    def list_cells(self, limit: int = 100, offset: int = 0,
                   agent_id: str = None, scene: str = None) -> List[dict]:
        pass

    @abstractmethod
    def delete_cell(self, cell_id: str) -> bool:
        pass

    @abstractmethod
    def put_scene(self, scene: dict) -> str:
        pass

    @abstractmethod
    def get_scene(self, scene_id: str) -> Optional[dict]:
        pass

    @abstractmethod
    def list_scenes(self, limit: int = 100, offset: int = 0, agent_id: str = None) -> List[dict]:
        pass

    @abstractmethod
    def put_skill(self, skill: dict) -> str:
        pass

    @abstractmethod
    def list_skills(self) -> List[dict]:
        pass

    @abstractmethod
    def close(self):
        pass



"""
Recall Engine - Pluggable retrieval architecture.

Supports multiple retrieval strategies and scene-based filtering.
Custom strategies can be injected via RecallStrategy interface.
"""

import re
from abc import ABC, abstractmethod
from typing import List
from dataclasses import dataclass, field

from openmemo.protocol.schemas import RecallResultItem


class RecallStrategy(ABC):
    @abstractmethod
    def retrieve(self, query: str, store=None, top_k: int = 10, **kwargs) -> List[RecallResultItem]:
        pass


class BM25Strategy(RecallStrategy):
    def __init__(self, config=None):
        from openmemo.config import RecallConfig
        self._config = config or RecallConfig()

    def retrieve(self, query: str, store=None, top_k: int = 10, **kwargs) -> List[RecallResultItem]:
        if not store:
            return []

        keywords = self._extract_keywords(query)
        if not keywords:
            return []

        agent_id = kwargs.get("agent_id")
        scene = kwargs.get("scene")
        all_cells = store.list_cells(agent_id=agent_id, scene=scene)
        scored = []

        for cell in all_cells:
            score = self._score(keywords, cell.get("content", ""))
            if score > 0:
                scored.append(RecallResultItem(
                    cell_id=cell.get("id", ""),
                    content=cell.get("content", ""),
                    score=score,
                    source="fast",
                ))

        scored.sort(key=lambda x: x.score, reverse=True)
        return scored[:top_k]

    def _extract_keywords(self, query: str) -> List[str]:
        stop_words = {"the", "a", "an", "is", "are", "was", "were", "in", "on", "at", "to", "for",
                      "of", "and", "or", "but", "not", "with", "this", "that", "it", "be", "have",
                      "do", "what", "how", "when", "where", "who", "which", "my", "your", "i"}
        words = re.findall(r'\w+', query.lower())
        return [w for w in words if w not in stop_words and len(w) > 1]

    def _score(self, keywords: List[str], text: str) -> float:
        text_lower = text.lower()
        score = 0.0
        for kw in keywords:
            if kw in text_lower:
                score += 1.0
        return score


class VectorStrategy(RecallStrategy):
    def __init__(self, vector_store=None, embed_fn=None):
        self._vector_store = vector_store
        self._embed_fn = embed_fn

    def retrieve(self, query: str, store=None, top_k: int = 10, **kwargs) -> List[RecallResultItem]:
        if not self._vector_store or not self._embed_fn:
            return []

        try:
            query_embedding = self._embed_fn(query)
            results = self._vector_store.search(query_embedding, top_k=top_k)
            return [
                RecallResultItem(
                    cell_id=r.get("id", ""),
                    content=r.get("content", ""),
                    score=r.get("score", 0.0),
                    source="middle",
                )
                for r in results
            ]
        except Exception:
            return []


@dataclass
class RecallResult:
    cell_id: str = ""
    content: str = ""
    score: float = 0.0
    source: str = "fast"
    metadata: dict = field(default_factory=dict)


class MergeStrategy(ABC):
    @abstractmethod
    def merge(self, results: List[RecallResultItem], top_k: int) -> List[RecallResult]:
        pass


class DefaultMergeStrategy(MergeStrategy):
    def merge(self, results: List[RecallResultItem], top_k: int) -> List[RecallResult]:
        seen = {}
        for r in results:
            if r.cell_id in seen:
                existing = seen[r.cell_id]
                existing.score = max(existing.score, r.score)
            else:
                seen[r.cell_id] = RecallResult(
                    cell_id=r.cell_id,
                    content=r.content,
                    score=r.score,
                    source=r.source,
                )

        merged = list(seen.values())
        merged.sort(key=lambda x: x.score, reverse=True)
        return merged[:top_k]


class RecallEngine:
    def __init__(self, store=None, vector_store=None, embed_fn=None,
                 strategies: List[RecallStrategy] = None,
                 merge_strategy: MergeStrategy = None,
                 config=None):
        from openmemo.config import RecallConfig
        self.store = store
        self._config = config or RecallConfig()

        if strategies is not None:
            self._strategies = strategies
        else:
            self._strategies = [BM25Strategy(config=self._config)]
            if vector_store and embed_fn:
                self._strategies.append(VectorStrategy(vector_store, embed_fn))

        self._merge_strategy = merge_strategy or DefaultMergeStrategy()

    def recall(self, query: str, top_k: int = 10, budget: int = 2000,
               agent_id: str = None, scene: str = None) -> List[RecallResult]:
        all_results = []

        for strategy in self._strategies:
            results = strategy.retrieve(
                query, store=self.store, top_k=top_k * 2,
                agent_id=agent_id, scene=scene,
            )
            all_results.extend(results)

        merged = self._merge_strategy.merge(all_results, top_k)
        return self._apply_budget(merged, budget)

    def _apply_budget(self, results: List[RecallResult], budget: int) -> List[RecallResult]:
        total_tokens = 0
        budgeted = []
        for r in results:
            tokens = len(r.content.split())
            if total_tokens + tokens > budget:
                break
            total_tokens += tokens
            budgeted.append(r)
        return budgeted



"""
OpenMemo - The Memory Infrastructure for AI Agents

Provides structured, evolving, and long-term memory
for autonomous AI systems.
"""

from openmemo.api.sdk import Memory, MemoryClient, OpenMemo
from openmemo.api.remote import RemoteMemory
from openmemo.config import OpenMemoConfig
from openmemo.core.memcell import CellType

__version__ = "0.6.0"
__all__ = ["OpenMemo", "Memory", "MemoryClient", "RemoteMemory", "OpenMemoConfig", "CellType"]



"""
MCP (Model Context Protocol) Adapter for OpenMemo.

Allows OpenMemo to serve as a memory backend for
Claude and other MCP-compatible AI assistants.

Supports both local SDK and remote API modes.

Usage (local):
    from openmemo.adapters.mcp import OpenMemoMCPServer
    server = OpenMemoMCPServer()

Usage (remote):
    from openmemo.adapters.mcp import OpenMemoMCPServer
    server = OpenMemoMCPServer(base_url="https://api.openmemo.ai")
"""


class OpenMemoMCPServer:
    def __init__(self, db_path: str = "openmemo.db", memory=None,
                 base_url: str = None, api_key: str = None):
        if memory:
            self.memory = memory
        elif base_url:
            from openmemo.api.remote import RemoteMemory
            self.memory = RemoteMemory(base_url=base_url, api_key=api_key)
        else:
            from openmemo.api.sdk import Memory
            self.memory = Memory(db_path=db_path)

    def get_tools(self):
        return [
            {
                "name": "write_memory",
                "description": "Store a memory. Use this to remember facts, decisions, preferences, or observations.",
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        "content": {"type": "string", "description": "The memory content to store"},
                        "scene": {"type": "string", "description": "Context scene (e.g., 'coding', 'deployment')"},
                        "memory_type": {"type": "string", "enum": ["fact", "decision", "preference", "constraint", "observation"]},
                        "confidence": {"type": "number", "description": "Confidence level 0-1", "default": 0.8},
                        "agent_id": {"type": "string", "description": "Agent identifier"},
                    },
                    "required": ["content"],
                },
            },
            {
                "name": "recall_context",
                "description": "Recall relevant memories for agent reasoning. Returns contextually relevant past memories.",
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        "query": {"type": "string", "description": "What to recall"},
                        "scene": {"type": "string", "description": "Filter by scene"},
                        "mode": {"type": "string", "enum": ["kv", "narrative", "raw"], "description": "Recall mode"},
                        "limit": {"type": "integer", "description": "Max results", "default": 5},
                        "agent_id": {"type": "string", "description": "Agent identifier"},
                    },
                    "required": ["query"],
                },
            },
            {
                "name": "search_memory",
                "description": "Search memories by query. Returns scored results.",
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        "query": {"type": "string", "description": "Search query"},
                        "scene": {"type": "string", "description": "Filter by scene"},
                        "limit": {"type": "integer", "description": "Max results", "default": 10},
                        "agent_id": {"type": "string", "description": "Agent identifier"},
                    },
                    "required": ["query"],
                },
            },
            {
                "name": "list_scenes",
                "description": "List all memory scenes.",
                "inputSchema": {
                    "type": "object",
                    "properties": {
                        "agent_id": {"type": "string", "description": "Agent identifier"},
                    },
                },
            },
        ]

    def handle_tool(self, tool_name: str, arguments: dict) -> dict:
        if tool_name == "write_memory":
            memory_id = self.memory.write_memory(
                content=arguments["content"],
                scene=arguments.get("scene", ""),
                memory_type=arguments.get("memory_type", "fact"),
                confidence=arguments.get("confidence", 0.8),
                agent_id=arguments.get("agent_id", ""),
            )
            return {"memory_id": memory_id, "status": "stored"}

        elif tool_name == "recall_context":
            return self.memory.recall_context(
                query=arguments["query"],
                scene=arguments.get("scene", ""),
                agent_id=arguments.get("agent_id", ""),
                mode=arguments.get("mode", "kv"),
                limit=arguments.get("limit", 5),
            )

        elif tool_name == "search_memory":
            results = self.memory.search_memory(
                query=arguments["query"],
                scene=arguments.get("scene", ""),
                agent_id=arguments.get("agent_id", ""),
                limit=arguments.get("limit", 10),
            )
            return {"results": results}

        elif tool_name == "list_scenes":
            scenes = self.memory.list_scenes(
                agent_id=arguments.get("agent_id", ""),
            )
            return {"scenes": scenes}

        return {"error": f"Unknown tool: {tool_name}"}

    def close(self):
        if hasattr(self.memory, 'close'):
            self.memory.close()



"""
SQLite storage backend - the default store.

Stores notes, MemCells, MemScenes, and skills in a local SQLite database.
Zero configuration, works out of the box.
"""

import json
import sqlite3
from typing import List, Optional
from openmemo.storage.base_store import BaseStore


class SQLiteStore(BaseStore):
    def __init__(self, db_path: str = "openmemo.db", check_same_thread: bool = True):
        self.db_path = db_path
        self.conn = sqlite3.connect(db_path, check_same_thread=check_same_thread)
        self.conn.row_factory = sqlite3.Row
        self._create_tables()

    def _create_tables(self):
        cursor = self.conn.cursor()
        cursor.executescript("""
            CREATE TABLE IF NOT EXISTS notes (
                id TEXT PRIMARY KEY,
                content TEXT NOT NULL,
                source TEXT DEFAULT 'manual',
                agent_id TEXT DEFAULT '',
                scene TEXT DEFAULT '',
                timestamp REAL,
                metadata TEXT DEFAULT '{}'
            );

            CREATE TABLE IF NOT EXISTS cells (
                id TEXT PRIMARY KEY,
                note_id TEXT,
                content TEXT NOT NULL,
                cell_type TEXT DEFAULT 'fact',
                facts TEXT DEFAULT '[]',
                stage TEXT DEFAULT 'exploration',
                importance REAL DEFAULT 0.5,
                access_count INTEGER DEFAULT 0,
                last_accessed REAL,
                created_at REAL,
                agent_id TEXT DEFAULT '',
                scene TEXT DEFAULT '',
                connections TEXT DEFAULT '[]',
                metadata TEXT DEFAULT '{}'
            );

            CREATE TABLE IF NOT EXISTS scenes (
                id TEXT PRIMARY KEY,
                title TEXT,
                summary TEXT,
                cell_ids TEXT DEFAULT '[]',
                theme TEXT,
                agent_id TEXT DEFAULT '',
                created_at REAL,
                updated_at REAL,
                metadata TEXT DEFAULT '{}'
            );

            CREATE TABLE IF NOT EXISTS skills (
                id TEXT PRIMARY KEY,
                name TEXT,
                description TEXT,
                pattern TEXT,
                usage_count INTEGER DEFAULT 0,
                success_rate REAL DEFAULT 0.0,
                created_at REAL,
                metadata TEXT DEFAULT '{}'
            );

            CREATE TABLE IF NOT EXISTS pyramid (
                id TEXT PRIMARY KEY,
                tier TEXT,
                content TEXT,
                source_ids TEXT DEFAULT '[]',
                created_at REAL,
                metadata TEXT DEFAULT '{}'
            );
        """)
        self.conn.commit()
        self._migrate()

    def _migrate(self):
        cursor = self.conn.cursor()
        for table, col, col_def in [
            ("notes", "agent_id", "TEXT DEFAULT ''"),
            ("notes", "scene", "TEXT DEFAULT ''"),
            ("cells", "agent_id", "TEXT DEFAULT ''"),
            ("cells", "scene", "TEXT DEFAULT ''"),
            ("cells", "cell_type", "TEXT DEFAULT 'fact'"),
            ("scenes", "agent_id", "TEXT DEFAULT ''"),
        ]:
            try:
                cursor.execute(f"ALTER TABLE {table} ADD COLUMN {col} {col_def}")
            except sqlite3.OperationalError:
                pass
        self.conn.commit()

    def put_note(self, note: dict) -> str:
        note_id = note.get("id", "")
        cursor = self.conn.cursor()
        cursor.execute(
            "INSERT OR REPLACE INTO notes (id, content, source, agent_id, scene, timestamp, metadata) VALUES (?, ?, ?, ?, ?, ?, ?)",
            (note_id, note.get("content", ""), note.get("source", "manual"),
             note.get("agent_id", ""), note.get("scene", ""),
             note.get("timestamp", 0), json.dumps(note.get("metadata", {})))
        )
        self.conn.commit()
        return note_id

    def get_note(self, note_id: str) -> Optional[dict]:
        cursor = self.conn.cursor()
        cursor.execute("SELECT * FROM notes WHERE id = ?", (note_id,))
        row = cursor.fetchone()
        if not row:
            return None
        return self._row_to_note(row)

    def list_notes(self, limit: int = 100, offset: int = 0, agent_id: str = None) -> List[dict]:
        cursor = self.conn.cursor()
        if agent_id:
            cursor.execute("SELECT * FROM notes WHERE agent_id = ? ORDER BY timestamp DESC LIMIT ? OFFSET ?",
                           (agent_id, limit, offset))
        else:
            cursor.execute("SELECT * FROM notes ORDER BY timestamp DESC LIMIT ? OFFSET ?", (limit, offset))
        return [self._row_to_note(row) for row in cursor.fetchall()]

    def delete_note(self, note_id: str) -> bool:
        cursor = self.conn.cursor()
        cursor.execute("DELETE FROM notes WHERE id = ?", (note_id,))
        self.conn.commit()
        return cursor.rowcount > 0

    def put_cell(self, cell: dict) -> str:
        cell_id = cell.get("id", "")
        cursor = self.conn.cursor()
        cursor.execute(
            """INSERT OR REPLACE INTO cells
            (id, note_id, content, cell_type, facts, stage, importance, access_count,
             last_accessed, created_at, agent_id, scene, connections, metadata)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)""",
            (cell_id, cell.get("note_id", ""), cell.get("content", ""),
             cell.get("cell_type", "fact"),
             json.dumps(cell.get("facts", [])), cell.get("stage", "exploration"),
             cell.get("importance", 0.5), cell.get("access_count", 0),
             cell.get("last_accessed", 0), cell.get("created_at", 0),
             cell.get("agent_id", ""), cell.get("scene", ""),
             json.dumps(cell.get("connections", [])), json.dumps(cell.get("metadata", {})))
        )
        self.conn.commit()
        return cell_id

    def get_cell(self, cell_id: str) -> Optional[dict]:
        cursor = self.conn.cursor()
        cursor.execute("SELECT * FROM cells WHERE id = ?", (cell_id,))
        row = cursor.fetchone()
        if not row:
            return None
        return self._row_to_cell(row)

    def list_cells(self, limit: int = 100, offset: int = 0,
                   agent_id: str = None, scene: str = None) -> List[dict]:
        cursor = self.conn.cursor()
        conditions = []
        params = []
        if agent_id:
            conditions.append("agent_id = ?")
            params.append(agent_id)
        if scene:
            conditions.append("scene = ?")
            params.append(scene)

        where = " WHERE " + " AND ".join(conditions) if conditions else ""
        params.extend([limit, offset])
        cursor.execute(f"SELECT * FROM cells{where} ORDER BY created_at DESC LIMIT ? OFFSET ?", params)
        return [self._row_to_cell(row) for row in cursor.fetchall()]

    def delete_cell(self, cell_id: str) -> bool:
        cursor = self.conn.cursor()
        cursor.execute("DELETE FROM cells WHERE id = ?", (cell_id,))
        self.conn.commit()
        return cursor.rowcount > 0

    def put_scene(self, scene: dict) -> str:
        scene_id = scene.get("id", "")
        cursor = self.conn.cursor()
        cursor.execute(
            """INSERT OR REPLACE INTO scenes
            (id, title, summary, cell_ids, theme, agent_id, created_at, updated_at, metadata)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)""",
            (scene_id, scene.get("title", ""), scene.get("summary", ""),
             json.dumps(scene.get("cell_ids", [])), scene.get("theme", ""),
             scene.get("agent_id", ""),
             scene.get("created_at", 0), scene.get("updated_at", 0),
             json.dumps(scene.get("metadata", {})))
        )
        self.conn.commit()
        return scene_id

    def get_scene(self, scene_id: str) -> Optional[dict]:
        cursor = self.conn.cursor()
        cursor.execute("SELECT * FROM scenes WHERE id = ?", (scene_id,))
        row = cursor.fetchone()
        if not row:
            return None
        return self._row_to_scene(row)

    def list_scenes(self, limit: int = 100, offset: int = 0, agent_id: str = None) -> List[dict]:
        cursor = self.conn.cursor()
        if agent_id:
            cursor.execute("SELECT * FROM scenes WHERE agent_id = ? ORDER BY updated_at DESC LIMIT ? OFFSET ?",
                           (agent_id, limit, offset))
        else:
            cursor.execute("SELECT * FROM scenes ORDER BY updated_at DESC LIMIT ? OFFSET ?", (limit, offset))
        return [self._row_to_scene(row) for row in cursor.fetchall()]

    def put_skill(self, skill: dict) -> str:
        skill_id = skill.get("id", "")
        cursor = self.conn.cursor()
        cursor.execute(
            """INSERT OR REPLACE INTO skills
            (id, name, description, pattern, usage_count, success_rate, created_at, metadata)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?)""",
            (skill_id, skill.get("name", ""), skill.get("description", ""),
             skill.get("pattern", ""), skill.get("usage_count", 0),
             skill.get("success_rate", 0.0), skill.get("created_at", 0),
             json.dumps(skill.get("metadata", {})))
        )
        self.conn.commit()
        return skill_id

    def list_skills(self) -> List[dict]:
        cursor = self.conn.cursor()
        cursor.execute("SELECT * FROM skills ORDER BY usage_count DESC")
        return [self._row_to_skill(row) for row in cursor.fetchall()]

    def close(self):
        self.conn.close()

    def _row_to_note(self, row) -> dict:
        d = {
            "id": row["id"], "content": row["content"], "source": row["source"],
            "timestamp": row["timestamp"], "metadata": json.loads(row["metadata"] or "{}"),
        }
        try:
            d["agent_id"] = row["agent_id"] or ""
            d["scene"] = row["scene"] or ""
        except (IndexError, KeyError):
            d["agent_id"] = ""
            d["scene"] = ""
        return d

    def _row_to_cell(self, row) -> dict:
        d = {
            "id": row["id"], "note_id": row["note_id"], "content": row["content"],
            "facts": json.loads(row["facts"] or "[]"), "stage": row["stage"],
            "importance": row["importance"], "access_count": row["access_count"],
            "last_accessed": row["last_accessed"], "created_at": row["created_at"],
            "connections": json.loads(row["connections"] or "[]"),
            "metadata": json.loads(row["metadata"] or "{}"),
        }
        try:
            d["agent_id"] = row["agent_id"] or ""
            d["scene"] = row["scene"] or ""
            d["cell_type"] = row["cell_type"] or "fact"
        except (IndexError, KeyError):
            d["agent_id"] = ""
            d["scene"] = ""
            d["cell_type"] = "fact"
        return d

    def _row_to_scene(self, row) -> dict:
        d = {
            "id": row["id"], "title": row["title"], "summary": row["summary"],
            "cell_ids": json.loads(row["cell_ids"] or "[]"), "theme": row["theme"],
            "created_at": row["created_at"], "updated_at": row["updated_at"],
            "metadata": json.loads(row["metadata"] or "{}"),
        }
        try:
            d["agent_id"] = row["agent_id"] or ""
        except (IndexError, KeyError):
            d["agent_id"] = ""
        return d

    def _row_to_skill(self, row) -> dict:
        return {
            "id": row["id"], "name": row["name"], "description": row["description"],
            "pattern": row["pattern"], "usage_count": row["usage_count"],
            "success_rate": row["success_rate"], "created_at": row["created_at"],
            "metadata": json.loads(row["metadata"] or "{}"),
        }



"""
OpenClaw Adapter for OpenMemo.

Provides a memory backend for OpenClaw agent sessions.
Supports both local SDK and remote API modes.

Usage (local):
    from openmemo.adapters.openclaw import OpenClawMemoryBackend
    backend = OpenClawMemoryBackend(agent_id="claw_agent")

Usage (remote):
    from openmemo.adapters.openclaw import OpenClawMemoryBackend
    backend = OpenClawMemoryBackend(
        agent_id="claw_agent",
        base_url="https://api.openmemo.ai",
    )
"""

from typing import List, Dict


class OpenClawMemoryBackend:
    def __init__(self, db_path: str = "openmemo.db", agent_id: str = "",
                 memory=None, base_url: str = None, api_key: str = None):
        self.agent_id = agent_id

        if memory:
            self._memory = memory
        elif base_url:
            from openmemo.api.remote import RemoteMemory
            self._memory = RemoteMemory(base_url=base_url, api_key=api_key)
        else:
            from openmemo.api.sdk import Memory
            self._memory = Memory(db_path=db_path)

    def write_memory(self, content: str, scene: str = "",
                     memory_type: str = "fact", confidence: float = 0.8) -> str:
        return self._memory.write_memory(
            content=content,
            scene=scene,
            memory_type=memory_type,
            confidence=confidence,
            agent_id=self.agent_id,
        )

    def recall_context(self, query: str, scene: str = "",
                       mode: str = "kv", limit: int = 5) -> dict:
        return self._memory.recall_context(
            query=query,
            scene=scene,
            agent_id=self.agent_id,
            mode=mode,
            limit=limit,
        )

    def search_memory(self, query: str, scene: str = "",
                      limit: int = 10) -> List[Dict]:
        return self._memory.search_memory(
            query=query,
            scene=scene,
            agent_id=self.agent_id,
            limit=limit,
        )

    def list_scenes(self) -> List[str]:
        return self._memory.list_scenes(agent_id=self.agent_id)

    def memory_governance(self, operation: str = "cleanup") -> dict:
        return self._memory.memory_governance(operation=operation)

    def write(self, content: str, scene: str = "", cell_type: str = "fact") -> str:
        return self.write_memory(content=content, scene=scene, memory_type=cell_type)

    def recall(self, query: str, scene: str = "", mode: str = "kv",
               limit: int = 10) -> dict:
        return self.recall_context(query=query, scene=scene, mode=mode, limit=limit)

    def search(self, query: str, limit: int = 10) -> List[Dict]:
        return self.search_memory(query=query, limit=limit)

    def get_context(self, query: str, scene: str = "", limit: int = 3) -> List[str]:
        result = self.recall_context(query=query, scene=scene, limit=limit)
        return result.get("context", [])

    def close(self):
        if hasattr(self._memory, 'close'):
            self._memory.close()



"""
OpenMemo SDK — Memory Core API v1.0

Usage:
    from openmemo import OpenMemo

    memo = OpenMemo()
    memo.write_memory("User prefers Python backend", scene="coding", memory_type="preference")
    result = memo.recall_context("programming language", scene="coding")
    print(result)  # {"context": ["User prefers Python backend"]}

Aliases:
    Memory, MemoryClient — same as OpenMemo
"""

from typing import List, Callable

from openmemo.config import OpenMemoConfig
from openmemo.core.memory import Note
from openmemo.core.memcell import MemCell, LifecycleStage
from openmemo.core.scene import MemScene
from openmemo.core.recall import RecallEngine
from openmemo.core.reconstruct import ReconstructiveRecall
from openmemo.storage.sqlite_store import SQLiteStore
from openmemo.storage.vector_store import VectorStore
from openmemo.pyramid.pyramid_engine import PyramidEngine
from openmemo.pyramid.summarizer import Summarizer
from openmemo.skill.skill_engine import SkillEngine
from openmemo.governance.conflict_detector import ConflictDetector
from openmemo.governance.version_manager import VersionManager


class Memory:
    """
    The main entry point for OpenMemo.

    Provides 5 core APIs:
        write_memory()       — Write a memory
        search_memory()      — Basic search
        recall_context()     — Agent reasoning context (most important)
        list_scenes()        — List memory scenes
        memory_governance()  — Memory cleanup & governance

    Args:
        db_path: Path to SQLite database file. Default: "openmemo.db"
        store: Custom storage backend (must implement BaseStore interface)
        embed_fn: Optional embedding function for vector search.
        config: Optional OpenMemoConfig for customizing engine behavior.
    """

    def __init__(self, db_path: str = "openmemo.db", store=None,
                 embed_fn: Callable = None, config: OpenMemoConfig = None):
        self._config = config or OpenMemoConfig()
        self.store = store or SQLiteStore(db_path)
        self.embed_fn = embed_fn
        self.vector_store = VectorStore() if embed_fn else None
        self.recall_engine = RecallEngine(
            store=self.store,
            vector_store=self.vector_store,
            embed_fn=self.embed_fn,
            config=self._config.recall,
        )
        self.reconstructor = ReconstructiveRecall(
            recall_engine=self.recall_engine,
            store=self.store,
        )
        self.pyramid = PyramidEngine(
            store=self.store,
            summarizer=Summarizer(),
            config=self._config.pyramid,
        )
        self.skill_engine = SkillEngine(
            store=self.store,
            config=self._config.skill,
        )
        self.conflict_detector = ConflictDetector(config=self._config.governance)
        self.version_manager = VersionManager()

    # ─── Core API 1: write_memory ───

    def write_memory(self, content: str, scene: str = "",
                     memory_type: str = "fact", confidence: float = 0.8,
                     agent_id: str = "", metadata: dict = None) -> str:
        return self._write_impl(
            content=content, scene=scene, cell_type=memory_type,
            confidence=confidence, agent_id=agent_id,
            source="manual", metadata=metadata,
        )

    # ─── Core API 2: search_memory ───

    def search_memory(self, query: str, scene: str = "",
                      agent_id: str = "", limit: int = 10) -> List[dict]:
        results = self.recall_engine.recall(
            query, top_k=limit, budget=50000,
            agent_id=agent_id or None, scene=scene or None,
        )
        return [
            {
                "content": r.content,
                "score": r.score,
                "cell_id": r.cell_id,
            }
            for r in results
        ]

    # ─── Core API 3: recall_context (most important) ───

    def recall_context(self, query: str, scene: str = "",
                       agent_id: str = "", limit: int = 5,
                       mode: str = "kv") -> dict:
        if mode == "narrative":
            result = self.reconstructor.reconstruct(
                query, max_sources=limit,
                agent_id=agent_id or None,
            )
            return {
                "memory_story": result.narrative,
                "sources": result.sources,
                "confidence": result.confidence,
            }

        if mode == "raw":
            return {
                "context": self._recall_raw_impl(
                    query, agent_id=agent_id, scene=scene,
                    top_k=limit,
                ),
            }

        results = self.recall_engine.recall(
            query, top_k=limit, budget=2000,
            agent_id=agent_id or None, scene=scene or None,
        )

        for r in results:
            cell = self.store.get_cell(r.cell_id)
            if cell:
                cell_obj = MemCell.from_dict(cell)
                cell_obj.access()
                self.store.put_cell(cell_obj.to_dict())

        return {
            "context": [r.content for r in results],
        }

    # ─── Core API 4: list_scenes ───

    def list_scenes(self, agent_id: str = "") -> List[str]:
        raw_scenes = self.store.list_scenes(agent_id=agent_id or None)
        return [s.get("title", "") for s in raw_scenes if s.get("title")]

    # ─── Core API 5: memory_governance ───

    def memory_governance(self, operation: str = "cleanup") -> dict:
        cells = self.store.list_cells(limit=500)

        if operation == "dedupe":
            return self._governance_dedupe(cells)
        elif operation == "merge":
            return self._governance_merge(cells)
        elif operation == "decay":
            return self._governance_decay(cells)
        elif operation == "cleanup":
            return self._governance_cleanup(cells)
        else:
            return {"error": f"Unknown operation: {operation}"}

    # ─── Backward-compatible aliases ───

    def write(self, content: str, agent_id: str = "", scene: str = "",
              cell_type: str = "fact", source: str = "manual",
              metadata: dict = None) -> str:
        return self._write_impl(
            content=content, scene=scene, cell_type=cell_type,
            confidence=0.8, agent_id=agent_id,
            source=source, metadata=metadata,
        )

    def add(self, content: str, source: str = "manual", agent_id: str = "",
            scene: str = "", cell_type: str = "fact", metadata: dict = None) -> str:
        return self._write_impl(
            content=content, scene=scene, cell_type=cell_type,
            confidence=0.8, agent_id=agent_id,
            source=source, metadata=metadata,
        )

    def recall(self, query: str, agent_id: str = "", scene: str = "",
               mode: str = "kv", top_k: int = None, limit: int = None,
               budget: int = 2000) -> dict:
        effective_limit = limit or top_k or 5
        return self.recall_context(
            query=query, scene=scene, agent_id=agent_id,
            limit=effective_limit, mode=mode,
        )

    def recall_raw(self, query: str, agent_id: str = "", scene: str = "",
                   top_k: int = 10, budget: int = 2000) -> List[dict]:
        return self._recall_raw_impl(query, agent_id, scene, top_k, budget)

    def search(self, query: str, agent_id: str = "",
               top_k: int = None, limit: int = None) -> List[dict]:
        effective_limit = limit or top_k or 10
        return self.search_memory(query, agent_id=agent_id, limit=effective_limit)

    def context(self, query: str, agent_id: str = "", scene: str = "",
                limit: int = 3) -> List[str]:
        result = self.recall_context(query, scene=scene, agent_id=agent_id, limit=limit)
        return result.get("context", [])

    def scenes(self, agent_id: str = "") -> List[str]:
        return self.list_scenes(agent_id=agent_id)

    def scenes_detail(self, agent_id: str = "") -> List[dict]:
        return self.store.list_scenes(agent_id=agent_id or None)

    def maintain(self) -> dict:
        return self.memory_governance("cleanup")

    def reconstruct(self, query: str, agent_id: str = "", max_sources: int = 10) -> dict:
        result = self.reconstructor.reconstruct(
            query, max_sources=max_sources,
            agent_id=agent_id or None,
        )
        return {
            "query": result.query,
            "narrative": result.narrative,
            "sources": result.sources,
            "confidence": result.confidence,
        }

    def delete(self, memory_id: str) -> bool:
        cell_deleted = self.store.delete_cell(memory_id)
        if cell_deleted:
            return True
        return self.store.delete_note(memory_id)

    def stats(self) -> dict:
        notes = self.store.list_notes(limit=10000)
        cells = self.store.list_cells(limit=10000)
        scenes_list = self.store.list_scenes(limit=10000)
        skills = self.store.list_skills()

        stage_counts = {}
        for c in cells:
            stage = c.get("stage", "unknown")
            stage_counts[stage] = stage_counts.get(stage, 0) + 1

        conflicts = self.conflict_detector.get_unresolved()

        return {
            "notes": len(notes),
            "cells": len(cells),
            "scenes": len(scenes_list),
            "skills": len(skills),
            "stages": stage_counts,
            "unresolved_conflicts": len(conflicts),
        }

    # ─── Internal implementations ───

    def _write_impl(self, content: str, scene: str, cell_type: str,
                    confidence: float, agent_id: str,
                    source: str, metadata: dict) -> str:
        from openmemo._internal import get_evolution_params

        note = Note(content=content, source=source, metadata=metadata or {})
        note_dict = note.to_dict()
        note_dict["agent_id"] = agent_id
        note_dict["scene"] = scene
        self.store.put_note(note_dict)

        importance = confidence if confidence else get_evolution_params()["default_importance"]
        cell = MemCell(
            note_id=note.id,
            content=content,
            cell_type=cell_type,
            stage=LifecycleStage.EXPLORATION,
            importance=importance,
            agent_id=agent_id,
            scene=scene,
        )
        cell.metadata["confidence"] = confidence

        existing_cells = self.store.list_cells(limit=50, agent_id=agent_id or None)
        conflicts = self.conflict_detector.detect(cell.to_dict(), existing_cells)
        if conflicts:
            cell.metadata["has_conflicts"] = True
            cell.metadata["conflict_count"] = len(conflicts)

        self.version_manager.snapshot(cell.to_dict(), change_type="create")
        self.store.put_cell(cell.to_dict())

        if scene:
            self._ensure_scene(scene, cell.id, agent_id)

        if self.vector_store and self.embed_fn:
            try:
                embedding = self.embed_fn(content)
                self.vector_store.add(cell.id, embedding, content)
            except Exception:
                pass

        return note.id

    def _recall_raw_impl(self, query: str, agent_id: str = "", scene: str = "",
                         top_k: int = 10, budget: int = 2000) -> List[dict]:
        results = self.recall_engine.recall(
            query, top_k=top_k, budget=budget,
            agent_id=agent_id or None, scene=scene or None,
        )

        for r in results:
            cell = self.store.get_cell(r.cell_id)
            if cell:
                cell_obj = MemCell.from_dict(cell)
                cell_obj.access()
                self.store.put_cell(cell_obj.to_dict())

        return [
            {
                "content": r.content,
                "score": r.score,
                "source": r.source,
                "cell_id": r.cell_id,
            }
            for r in results
        ]

    def _ensure_scene(self, scene_name: str, cell_id: str, agent_id: str = ""):
        all_scenes = self.store.list_scenes(agent_id=agent_id or None)
        for s in all_scenes:
            if s.get("title") == scene_name:
                scene_obj = MemScene.from_dict(s)
                scene_obj.add_cell(cell_id)
                self.store.put_scene(scene_obj.to_dict())
                return

        new_scene = MemScene(title=scene_name, theme=scene_name)
        new_scene.add_cell(cell_id)
        scene_dict = new_scene.to_dict()
        scene_dict["agent_id"] = agent_id
        self.store.put_scene(scene_dict)

    def _governance_dedupe(self, cells: list) -> dict:
        seen = {}
        duplicates = []
        for c in cells:
            content = c.get("content", "").strip().lower()
            if content in seen:
                duplicates.append(c["id"])
                self.store.delete_cell(c["id"])
            else:
                seen[content] = c["id"]
        return {
            "operation": "dedupe",
            "duplicates_removed": len(duplicates),
            "total_cells": len(cells),
        }

    def _governance_merge(self, cells: list) -> dict:
        pyramid_result = self.pyramid.process(cells)
        return {
            "operation": "merge",
            "pyramid": pyramid_result,
            "total_cells": len(cells),
        }

    def _governance_decay(self, cells: list) -> dict:
        import time
        now = time.time()
        decayed = 0
        for c in cells:
            last_access = c.get("last_access", c.get("created_at", now))
            age_days = (now - last_access) / 86400
            if age_days > 30:
                cell_obj = MemCell.from_dict(c)
                cell_obj.importance = max(0.1, cell_obj.importance * 0.9)
                self.store.put_cell(cell_obj.to_dict())
                decayed += 1
        return {
            "operation": "decay",
            "decayed_cells": decayed,
            "total_cells": len(cells),
        }

    def _governance_cleanup(self, cells: list) -> dict:
        dedupe_result = self._governance_dedupe(cells)
        remaining_cells = self.store.list_cells(limit=500)
        merge_result = self._governance_merge(remaining_cells)
        skills = self.skill_engine.extract_skills()
        return {
            "operation": "cleanup",
            "duplicates_removed": dedupe_result["duplicates_removed"],
            "pyramid": merge_result["pyramid"],
            "new_skills": len(skills),
            "total_cells": len(remaining_cells),
        }

    def close(self):
        self.store.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


OpenMemo = Memory
MemoryClient = Memory



"""
OpenMemo Configuration - Centralized configuration management.

Public configuration exposes only high-level behavioral switches.
Internal tuning parameters are encapsulated within engine
implementations and not exposed through this interface.
"""

from dataclasses import dataclass, field
from typing import Dict, Any


@dataclass
class RecallConfig:
    fast_brain_enabled: bool = True
    middle_brain_enabled: bool = True
    default_top_k: int = 10
    default_budget: int = 2000


@dataclass
class GovernanceConfig:
    conflict_detection_enabled: bool = True


@dataclass
class PyramidConfig:
    enabled: bool = True


@dataclass
class SkillConfig:
    auto_extract: bool = True


@dataclass
class EvolutionConfig:
    enabled: bool = True


@dataclass
class OpenMemoConfig:
    recall: RecallConfig = field(default_factory=RecallConfig)
    governance: GovernanceConfig = field(default_factory=GovernanceConfig)
    evolution: EvolutionConfig = field(default_factory=EvolutionConfig)
    pyramid: PyramidConfig = field(default_factory=PyramidConfig)
    skill: SkillConfig = field(default_factory=SkillConfig)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "OpenMemoConfig":
        config = cls()
        section_map = {
            "recall": config.recall,
            "governance": config.governance,
            "evolution": config.evolution,
            "pyramid": config.pyramid,
            "skill": config.skill,
        }
        for section_name, section_obj in section_map.items():
            if section_name in data:
                for k, v in data[section_name].items():
                    if hasattr(section_obj, k):
                        setattr(section_obj, k, v)
        return config



"""
OpenMemo REST Server — Memory Core API v1.0

Agent-First HTTP API for memory operations.
Install with: pip install openmemo[server]

Endpoints:
    POST /memory/write       — Write a memory
    POST /memory/search      — Basic search
    POST /memory/recall      — Recall context (kv/narrative/raw)
    GET  /memory/scenes      — List scenes
    POST /memory/governance  — Memory governance operations
    DELETE /memory/{id}      — Delete a memory
    POST /agent/context      — Get context for prompt injection
    GET  /version            — Get version info
"""

import os

try:
    from flask import Flask, request, jsonify, make_response
    from flask_cors import CORS
except ImportError:
    raise ImportError("Install server dependencies: pip install openmemo[server]")

from openmemo.api.sdk import Memory
from openmemo.api.docs import API_DOCS_HTML
from openmemo.api.inspector_html import INSPECTOR_HTML
from openmemo.config import OpenMemoConfig

API_VERSION = "1.0"
ENGINE_VERSION = "0.6.0"
SCHEMA_VERSION = "2"
CORE_VERSION = "0.6.0"
ADAPTER_VERSION = "2.2.0"


def create_app(db_path: str = None, config: OpenMemoConfig = None) -> Flask:
    app = Flask(__name__)
    CORS(app)

    db = db_path or os.environ.get("OPENMEMO_DB", "openmemo.db")
    from openmemo.storage.sqlite_store import SQLiteStore
    store = SQLiteStore(db_path=db, check_same_thread=False)
    memory = Memory(db_path=db, store=store, config=config)

    @app.route("/")
    def index():
        return jsonify({
            "service": "OpenMemo Memory API",
            "api_version": API_VERSION,
            "engine_version": ENGINE_VERSION,
            "description": "The Memory Infrastructure for AI Agents",
            "status": "running",
            "docs": "/docs",
            "github": "https://github.com/openmemoai/openmemo",
            "endpoints": {
                "write": "POST /memory/write",
                "search": "POST /memory/search",
                "recall": "POST /memory/recall",
                "scenes": "GET /memory/scenes",
                "governance": "POST /memory/governance",
                "delete": "DELETE /memory/{id}",
                "context": "POST /agent/context",
                "health": "GET /health",
                "version": "GET /version",
                "stats": "GET /api/stats",
            },
        })

    @app.route("/docs")
    def docs():
        response = make_response(API_DOCS_HTML)
        response.headers["Content-Type"] = "text/html"
        return response

    @app.route("/version")
    def version():
        return jsonify({
            "latest_core": CORE_VERSION,
            "latest_adapter": ADAPTER_VERSION,
            "schema_version": SCHEMA_VERSION,
        })

    @app.route("/health")
    def health():
        return jsonify({
            "status": "ok",
            "service": "openmemo",
            "api_version": API_VERSION,
            "engine_version": ENGINE_VERSION,
        })

    @app.route("/memory/write", methods=["POST"])
    @app.route("/api/memories", methods=["POST"])
    def write_memory():
        data = request.get_json()
        if not data or "content" not in data:
            return jsonify({"error": "content is required"}), 400

        memory_type = data.get("type", data.get("memory_type", data.get("cell_type", "fact")))
        memory_id = memory.write_memory(
            content=data["content"],
            scene=data.get("scene", ""),
            memory_type=memory_type,
            confidence=data.get("confidence", 0.8),
            agent_id=data.get("agent_id", ""),
            metadata=data.get("metadata", {}),
        )
        return jsonify({"memory_id": memory_id, "status": "stored"}), 201

    @app.route("/memory/search", methods=["POST"])
    @app.route("/api/memories/search", methods=["POST"])
    def search_mem():
        data = request.get_json()
        if not data or "query" not in data:
            return jsonify({"error": "query is required"}), 400

        results = memory.search_memory(
            query=data["query"],
            scene=data.get("scene", ""),
            agent_id=data.get("agent_id", ""),
            limit=data.get("limit", 10),
        )
        return jsonify({"results": results})

    @app.route("/memory/recall", methods=["POST"])
    @app.route("/api/memories/recall", methods=["POST"])
    def recall_mem():
        data = request.get_json()
        if not data or "query" not in data:
            return jsonify({"error": "query is required"}), 400

        result = memory.recall_context(
            query=data["query"],
            scene=data.get("scene", ""),
            agent_id=data.get("agent_id", ""),
            limit=data.get("limit", 5),
            mode=data.get("mode", "kv"),
        )
        return jsonify(result)

    @app.route("/memory/scenes", methods=["GET"])
    def scenes():
        agent_id = request.args.get("agent_id", "")
        scene_names = memory.list_scenes(agent_id=agent_id)
        return jsonify({"scenes": scene_names})

    @app.route("/memory/governance", methods=["POST"])
    @app.route("/api/maintain", methods=["POST"])
    def governance():
        data = request.get_json() or {}
        operation = data.get("operation", "cleanup")
        result = memory.memory_governance(operation=operation)
        return jsonify(result)

    @app.route("/memory/<memory_id>", methods=["DELETE"])
    def delete_memory(memory_id):
        deleted = memory.delete(memory_id)
        if deleted:
            return jsonify({"deleted": True}), 200
        return jsonify({"error": "not found"}), 404

    @app.route("/agent/context", methods=["POST"])
    def agent_context():
        data = request.get_json()
        if not data or "query" not in data:
            return jsonify({"error": "query is required"}), 400

        result = memory.recall_context(
            query=data["query"],
            scene=data.get("scene", ""),
            agent_id=data.get("agent_id", ""),
            limit=data.get("limit", 3),
            mode="kv",
        )
        return jsonify({"memory_context": result.get("context", [])})

    @app.route("/memory/reconstruct", methods=["POST"])
    @app.route("/api/memories/reconstruct", methods=["POST"])
    def reconstruct_memory():
        data = request.get_json()
        if not data or "query" not in data:
            return jsonify({"error": "query is required"}), 400

        result = memory.reconstruct(
            query=data["query"],
            agent_id=data.get("agent_id", ""),
            max_sources=data.get("max_sources", 10),
        )
        return jsonify(result)

    @app.route("/api/stats")
    def api_stats():
        return jsonify(memory.stats())

    @app.route("/inspector")
    def inspector():
        response = make_response(INSPECTOR_HTML)
        response.headers["Content-Type"] = "text/html"
        return response

    @app.route("/api/inspector/checklist")
    def inspector_checklist():
        checks = []

        checks.append({"name": "Adapter Loaded", "status": "ok"})

        try:
            memory.stats()
            checks.append({"name": "Memory Backend Connected", "status": "ok"})
        except Exception:
            checks.append({"name": "Memory Backend Connected", "status": "fail"})

        checks.append({"name": "Pipeline Active", "status": "ok"})

        try:
            memory.write_memory(content="__healthcheck__", scene="__healthcheck__",
                                memory_type="fact", confidence=0.1, agent_id="__system__")
            memory.memory_governance(operation="cleanup")
            checks.append({"name": "Memory Write Pipeline Healthy", "status": "ok"})
        except Exception:
            checks.append({"name": "Memory Write Pipeline Healthy", "status": "fail"})

        try:
            result = memory.search_memory(query="test", limit=1)
            if isinstance(result, list) and len(result) == 0:
                checks.append({"name": "Memory Recall Working", "status": "cold_start"})
            else:
                checks.append({"name": "Memory Recall Working", "status": "ok"})
        except Exception:
            checks.append({"name": "Memory Recall Working", "status": "fail"})

        checks.append({"name": "Task Memory & Deduplication Active", "status": "ok"})

        return jsonify({"checks": checks})

    @app.route("/api/inspector/memory-summary")
    def inspector_memory_summary():
        s = memory.stats()
        notes = memory.store.list_notes(limit=10000)
        cells = memory.store.list_cells(limit=10000)
        scenes_list = memory.list_scenes()

        type_dist = {}
        for c in cells:
            ct = c.get("cell_type", c.get("type", "unknown"))
            type_dist[ct] = type_dist.get(ct, 0) + 1

        scene_dist = {}
        for n in notes:
            sc = n.get("scene", "") or ""
            scene_dist[sc] = scene_dist.get(sc, 0) + 1

        return jsonify({
            "total_memories": s.get("notes", 0),
            "total_cells": s.get("cells", 0),
            "total_scenes": s.get("scenes", 0),
            "type_distribution": type_dist,
            "scene_distribution": scene_dist,
        })

    @app.route("/api/inspector/recent")
    def inspector_recent():
        limit = request.args.get("limit", 10, type=int)
        notes = memory.store.list_notes(limit=limit)
        recent = []
        for n in reversed(notes[-limit:]):
            recent.append({
                "content": n.get("content", ""),
                "scene": n.get("scene", ""),
                "memory_type": n.get("memory_type", n.get("type", "")),
                "timestamp": n.get("created_at", ""),
            })
        return jsonify({"recent": recent})

    @app.route("/api/inspector/search")
    def inspector_search():
        q = request.args.get("q", "")
        if not q:
            return jsonify({"results": []})
        limit = request.args.get("limit", 10, type=int)
        results = memory.search_memory(query=q, limit=limit)
        return jsonify({"results": results})

    @app.route("/api/inspector/health")
    def inspector_health():
        s = memory.stats()
        return jsonify({
            "status": "ok",
            "backend": "openmemo",
            "api_version": API_VERSION,
            "engine_version": ENGINE_VERSION,
            "total_memories": s.get("notes", 0),
            "total_scenes": s.get("scenes", 0),
        })

    return app


if __name__ == "__main__":
    app = create_app()
    port = int(os.environ.get("PORT", 8080))
    app.run(host="0.0.0.0", port=port)



"""
LangChain Adapter for OpenMemo.

Provides OpenMemoMemory() that works as a LangChain
BaseMemory compatible memory backend.

Supports both local SDK and remote API modes.

Usage (local):
    from openmemo.adapters.langchain import OpenMemoMemory
    memory = OpenMemoMemory(agent_id="my_agent")

Usage (remote):
    from openmemo.adapters.langchain import OpenMemoMemory
    memory = OpenMemoMemory(
        agent_id="my_agent",
        base_url="https://api.openmemo.ai",
    )
"""

from typing import Any, Dict, List


class OpenMemoMemory:
    memory_key: str = "history"

    def __init__(self, db_path: str = "openmemo.db", agent_id: str = "",
                 memory=None, memory_key: str = "history",
                 base_url: str = None, api_key: str = None):
        self.agent_id = agent_id
        self.memory_key = memory_key

        if memory:
            self._memory = memory
        elif base_url:
            from openmemo.api.remote import RemoteMemory
            self._memory = RemoteMemory(base_url=base_url, api_key=api_key)
        else:
            from openmemo.api.sdk import Memory
            self._memory = Memory(db_path=db_path)

    @property
    def memory_variables(self) -> List[str]:
        return [self.memory_key]

    def load_memory_variables(self, inputs: Dict[str, Any]) -> Dict[str, str]:
        query = inputs.get("input", "")
        if not query:
            return {self.memory_key: ""}

        result = self._memory.recall_context(
            query=query,
            agent_id=self.agent_id,
            limit=5,
            mode="kv",
        )

        context = result.get("context", [])
        if not context:
            return {self.memory_key: ""}

        return {self.memory_key: "\n".join(context)}

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        user_input = inputs.get("input", "")
        ai_output = outputs.get("output", "")

        if user_input:
            self._memory.write_memory(
                content=f"User: {user_input}",
                scene="conversation",
                memory_type="observation",
                agent_id=self.agent_id,
            )

        if ai_output:
            self._memory.write_memory(
                content=f"Assistant: {ai_output}",
                scene="conversation",
                memory_type="observation",
                agent_id=self.agent_id,
            )

    def clear(self) -> None:
        pass

    def close(self):
        if hasattr(self._memory, 'close'):
            self._memory.close()



"""
Skill Engine - Experience to skill extraction.

Detects repeated patterns in agent behavior and extracts
them into reusable skills. Extraction and relevance scoring
are pluggable via abstract base classes.
"""

import uuid
import time
from abc import ABC, abstractmethod
from typing import List
from dataclasses import dataclass, field


@dataclass
class Skill:
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    description: str = ""
    pattern: str = ""
    usage_count: int = 0
    success_rate: float = 0.0
    created_at: float = field(default_factory=time.time)
    metadata: dict = field(default_factory=dict)

    def record_usage(self, success: bool):
        self.usage_count += 1
        total_success = self.success_rate * (self.usage_count - 1) + (1.0 if success else 0.0)
        self.success_rate = total_success / self.usage_count

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "pattern": self.pattern,
            "usage_count": self.usage_count,
            "success_rate": self.success_rate,
            "created_at": self.created_at,
            "metadata": self.metadata,
        }


class SkillExtractor(ABC):
    @abstractmethod
    def should_extract(self, pattern: str, count: int, successes: int) -> bool:
        pass


class DefaultSkillExtractor(SkillExtractor):
    def __init__(self, config=None):
        self._config = config

    def should_extract(self, pattern: str, count: int, successes: int) -> bool:
        from openmemo._internal import get_skill_params
        params = get_skill_params()
        return count >= params["pattern_threshold"]


class RelevanceScorer(ABC):
    @abstractmethod
    def score(self, skill: dict, context: str) -> float:
        pass


class DefaultRelevanceScorer(RelevanceScorer):
    def score(self, skill: dict, context: str) -> float:
        pattern = skill.get("pattern", "").lower()
        context_lower = context.lower()
        if pattern and pattern in context_lower:
            return skill.get("success_rate", 0.0)
        return 0.0


class SkillEngine:
    def __init__(self, store=None, extractor: SkillExtractor = None,
                 scorer: RelevanceScorer = None, config=None):
        self.store = store
        self._extractor = extractor or DefaultSkillExtractor(config=config)
        self._scorer = scorer or DefaultRelevanceScorer()
        self._pattern_counts = {}

    def observe(self, action: str, context: str = "", success: bool = True):
        key = self._normalize(action)
        if key not in self._pattern_counts:
            self._pattern_counts[key] = {"count": 0, "successes": 0, "context": context}

        self._pattern_counts[key]["count"] += 1
        if success:
            self._pattern_counts[key]["successes"] += 1

    def extract_skills(self) -> List[Skill]:
        new_skills = []

        for pattern, data in self._pattern_counts.items():
            if self._extractor.should_extract(pattern, data["count"], data["successes"]):
                skill = Skill(
                    name=pattern,
                    description=f"Learned from {data['count']} observations",
                    pattern=pattern,
                    usage_count=data["count"],
                    success_rate=data["successes"] / data["count"] if data["count"] > 0 else 0,
                )
                new_skills.append(skill)

                if self.store:
                    self.store.put_skill(skill.to_dict())

        return new_skills

    def get_relevant_skills(self, context: str, top_k: int = 5) -> List[Skill]:
        if not self.store:
            return []

        all_skills = self.store.list_skills()

        scored = []
        for s in all_skills:
            relevance = self._scorer.score(s, context)
            if relevance > 0:
                scored.append((Skill(**{k: v for k, v in s.items() if k in Skill.__dataclass_fields__}), relevance))

        scored.sort(key=lambda x: x[1], reverse=True)
        return [s for s, _ in scored[:top_k]]

    def _normalize(self, action: str) -> str:
        return action.strip().lower()



"""
Memory Inspector Dashboard — HTML/CSS/JS for the Inspector Web UI.
"""

INSPECTOR_HTML = """<!DOCTYPE html>
<html lang="en">
<head>
<meta charset="UTF-8">
<meta name="viewport" content="width=device-width, initial-scale=1.0">
<title>OpenMemo Memory Inspector</title>
<style>
* { margin: 0; padding: 0; box-sizing: border-box; }
body { font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', sans-serif; background: #0d1117; color: #c9d1d9; }
.header { background: #161b22; border-bottom: 1px solid #30363d; padding: 16px 24px; display: flex; align-items: center; justify-content: space-between; }
.header h1 { font-size: 20px; color: #58a6ff; }
.header .version { font-size: 13px; color: #8b949e; }
.container { max-width: 1200px; margin: 0 auto; padding: 24px; display: grid; grid-template-columns: 1fr 1fr; gap: 16px; }
.card { background: #161b22; border: 1px solid #30363d; border-radius: 8px; padding: 20px; }
.card.full { grid-column: 1 / -1; }
.card h2 { font-size: 15px; color: #f0f6fc; margin-bottom: 12px; display: flex; align-items: center; gap: 8px; }
.card h2 .icon { font-size: 16px; }
.checklist-item { display: flex; align-items: center; gap: 8px; padding: 6px 0; font-size: 14px; }
.status-ok { color: #3fb950; }
.status-warn { color: #d29922; }
.status-fail { color: #f85149; }
.status-cold { color: #8b949e; }
.stat-row { display: flex; justify-content: space-between; padding: 6px 0; font-size: 14px; border-bottom: 1px solid #21262d; }
.stat-row:last-child { border-bottom: none; }
.stat-value { color: #58a6ff; font-weight: 600; }
.dist-bar { display: flex; align-items: center; gap: 8px; padding: 4px 0; font-size: 13px; }
.dist-bar .bar { height: 8px; border-radius: 4px; background: #58a6ff; min-width: 4px; }
.dist-bar .label { min-width: 100px; color: #8b949e; }
.dist-bar .count { color: #c9d1d9; min-width: 30px; text-align: right; }
.memory-item { padding: 10px 0; border-bottom: 1px solid #21262d; font-size: 14px; }
.memory-item:last-child { border-bottom: none; }
.memory-item .content { color: #c9d1d9; margin-bottom: 4px; }
.memory-item .meta { font-size: 12px; color: #8b949e; display: flex; gap: 12px; }
.tag { display: inline-block; padding: 2px 8px; border-radius: 12px; font-size: 11px; background: #21262d; color: #8b949e; }
.tag.scene { background: #0d419d; color: #a5d6ff; }
.tag.type { background: #1a4731; color: #56d364; }
.search-box { width: 100%; padding: 10px 14px; background: #0d1117; border: 1px solid #30363d; border-radius: 6px; color: #c9d1d9; font-size: 14px; outline: none; }
.search-box:focus { border-color: #58a6ff; }
.search-results { margin-top: 12px; }
.badge { display: inline-block; padding: 2px 10px; border-radius: 12px; font-size: 12px; font-weight: 600; }
.badge.ok { background: #1a4731; color: #3fb950; }
.badge.update { background: #3d2e00; color: #d29922; }
.refresh-note { font-size: 12px; color: #484f58; }
.empty { color: #484f58; font-size: 13px; font-style: italic; padding: 8px 0; }
.scene-badge { display: inline-block; padding: 4px 12px; background: #0d419d; color: #a5d6ff; border-radius: 16px; font-size: 14px; font-weight: 600; }
</style>
</head>
<body>
<div class="header">
  <h1>OpenMemo Memory Inspector</h1>
  <div>
    <span class="version" id="header-version"></span>
    <span class="refresh-note"> &middot; auto-refresh 5s</span>
  </div>
</div>

<div class="container">
  <!-- Checklist -->
  <div class="card">
    <h2><span class="icon">🔍</span> Memory System Checklist</h2>
    <div id="checklist"></div>
  </div>

  <!-- System Status -->
  <div class="card">
    <h2><span class="icon">⚙️</span> System Status</h2>
    <div id="system-status"></div>
  </div>

  <!-- Memory Records -->
  <div class="card">
    <h2><span class="icon">📊</span> Memory Records</h2>
    <div id="memory-records"></div>
  </div>

  <!-- Scene & Type Distribution -->
  <div class="card">
    <h2><span class="icon">🏷️</span> Distribution</h2>
    <div id="distribution"></div>
  </div>

  <!-- Recent Writes -->
  <div class="card full">
    <h2><span class="icon">📝</span> Recent Writes</h2>
    <div id="recent-writes"></div>
  </div>

  <!-- Memory Search -->
  <div class="card full">
    <h2><span class="icon">🔎</span> Memory Search</h2>
    <input type="text" class="search-box" id="search-input" placeholder="Search memories..." />
    <div class="search-results" id="search-results"></div>
  </div>

  <!-- Update Status -->
  <div class="card">
    <h2><span class="icon">📦</span> Update Status</h2>
    <div id="update-status"></div>
  </div>
</div>

<script>
const API = '';

function statusIcon(s) {
  if (s === 'ok') return '<span class="status-ok">✓</span>';
  if (s === 'warning') return '<span class="status-warn">!</span>';
  if (s === 'cold_start') return '<span class="status-cold">○</span>';
  return '<span class="status-fail">✕</span>';
}

async function fetchJSON(url) {
  try {
    const r = await fetch(API + url);
    return await r.json();
  } catch (e) { return null; }
}

async function loadChecklist() {
  const d = await fetchJSON('/api/inspector/checklist');
  if (!d) { document.getElementById('checklist').innerHTML = '<div class="empty">Could not load checklist</div>'; return; }
  const el = document.getElementById('checklist');
  el.innerHTML = d.checks.map(c =>
    '<div class="checklist-item">' + statusIcon(c.status) + ' ' + c.name + '</div>'
  ).join('');
}

async function loadHealth() {
  const d = await fetchJSON('/api/inspector/health');
  if (!d) return;
  const el = document.getElementById('system-status');
  el.innerHTML = [
    ['Memory Backend', d.backend || 'openmemo'],
    ['API Version', d.api_version || '-'],
    ['Engine Version', d.engine_version || '-'],
    ['Status', d.status || '-'],
    ['Total Memories', d.total_memories || 0],
    ['Total Scenes', d.total_scenes || 0],
  ].map(([k,v]) => '<div class="stat-row"><span>' + k + '</span><span class="stat-value">' + v + '</span></div>').join('');
}

async function loadSummary() {
  const d = await fetchJSON('/api/inspector/memory-summary');
  if (!d) return;

  document.getElementById('memory-records').innerHTML = [
    ['Total Memories', d.total_memories],
    ['Total Cells', d.total_cells],
    ['Scenes', d.total_scenes],
  ].map(([k,v]) => '<div class="stat-row"><span>' + k + '</span><span class="stat-value">' + v + '</span></div>').join('');

  let distHtml = '';
  if (d.type_distribution && Object.keys(d.type_distribution).length > 0) {
    const maxT = Math.max(...Object.values(d.type_distribution), 1);
    distHtml += '<div style="margin-bottom:12px"><div style="font-size:13px;color:#8b949e;margin-bottom:6px">By Type</div>';
    for (const [k,v] of Object.entries(d.type_distribution)) {
      const w = Math.max(4, (v/maxT)*100);
      distHtml += '<div class="dist-bar"><span class="label">' + k + '</span><div class="bar" style="width:' + w + 'px"></div><span class="count">' + v + '</span></div>';
    }
    distHtml += '</div>';
  }
  if (d.scene_distribution && Object.keys(d.scene_distribution).length > 0) {
    const maxS = Math.max(...Object.values(d.scene_distribution), 1);
    distHtml += '<div><div style="font-size:13px;color:#8b949e;margin-bottom:6px">By Scene</div>';
    for (const [k,v] of Object.entries(d.scene_distribution)) {
      const w = Math.max(4, (v/maxS)*100);
      distHtml += '<div class="dist-bar"><span class="label">' + (k||'(none)') + '</span><div class="bar" style="width:' + w + 'px;background:#3fb950"></div><span class="count">' + v + '</span></div>';
    }
    distHtml += '</div>';
  }
  document.getElementById('distribution').innerHTML = distHtml || '<div class="empty">No distribution data</div>';
}

function renderMemory(m) {
  const content = m.content || m.text || '';
  const scene = m.scene || '';
  const mtype = m.memory_type || m.cell_type || m.type || '';
  const score = m.score != null ? m.score.toFixed(2) : '';
  let meta = '';
  if (scene) meta += '<span class="tag scene">' + scene + '</span>';
  if (mtype) meta += '<span class="tag type">' + mtype + '</span>';
  if (score) meta += '<span class="tag">score: ' + score + '</span>';
  return '<div class="memory-item"><div class="content">' + escapeHtml(content.substring(0, 200)) + '</div><div class="meta">' + meta + '</div></div>';
}

function escapeHtml(t) {
  const d = document.createElement('div'); d.textContent = t; return d.innerHTML;
}

async function loadRecent() {
  const d = await fetchJSON('/api/inspector/recent');
  const el = document.getElementById('recent-writes');
  if (!d || !d.recent || d.recent.length === 0) { el.innerHTML = '<div class="empty">No recent writes (cold start)</div>'; return; }
  el.innerHTML = d.recent.map(renderMemory).join('');
}

async function loadVersion() {
  const d = await fetchJSON('/version');
  if (!d) return;
  const el = document.getElementById('update-status');
  el.innerHTML = [
    ['OpenMemo Core', d.latest_core],
    ['OpenMemo Adapter', d.latest_adapter],
    ['Schema Version', d.schema_version],
  ].map(([k,v]) => '<div class="stat-row"><span>' + k + '</span><span class="stat-value">' + v + '</span></div>').join('');
  document.getElementById('header-version').textContent = 'Core ' + d.latest_core + ' / Adapter ' + d.latest_adapter;
}

let searchTimer;
document.getElementById('search-input').addEventListener('input', function() {
  clearTimeout(searchTimer);
  const q = this.value.trim();
  if (q.length < 2) { document.getElementById('search-results').innerHTML = ''; return; }
  searchTimer = setTimeout(() => doSearch(q), 300);
});

async function doSearch(q) {
  const d = await fetchJSON('/api/inspector/search?q=' + encodeURIComponent(q));
  const el = document.getElementById('search-results');
  if (!d || !d.results || d.results.length === 0) { el.innerHTML = '<div class="empty">No results for "' + escapeHtml(q) + '"</div>'; return; }
  el.innerHTML = d.results.map(renderMemory).join('');
}

async function refreshAll() {
  await Promise.all([loadChecklist(), loadHealth(), loadSummary(), loadRecent(), loadVersion()]);
}

refreshAll();
setInterval(refreshAll, 5000);
</script>
</body>
</html>"""
